/****************************************************************************/
/*		SI4MPS, I Kolokvijum, Zadatak 5., 29.10.2009.						*/
/*      Autor resenja: Ivan Kuraj 06/0398                                   */
/****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <math.h>
#ifdef __GNUC__
#include <stdint.h>
#endif

/* NUMERICAL CONSTANTS ******************************************************/

/* broj stvorenih niti */
#define NUM_THREADS 5
/* maksimalan broj elemenata niza */
#define MAX_COUNT 1000

/* GLOBAL DATA **************************************************************/

/* jedinstvena brava za sve niti i njihove podatke */
pthread_mutex_t mutex;
/* indikator i odgovarajuca uslovna promenljiva za cekanje da sve niti budu
 * kreirane */
pthread_cond_t thread_start_cond;
int thread_creation_finished;
/* indikator i odgovarajuca uslovna promenljiva za signalizaciju zavrsetka
 * rada jedne niti, kao i cekanje na ispis glavne niti */
pthread_cond_t thread_finished_cond, thread_finish_wait_cond;
int thread_finished;
/* indikator da li je nit promenila max vrednost */
int max_changed;

/* globalne promenljive niza, maksimuma, velicine niza i broja elemenata za
 * obradu svake niti */
double array[MAX_COUNT], max;
int n, array_slice;

/* AUXILIARY FUNCTIONS (DECLARATIONS) ***************************************/

/* funkcija cita sa ulaza broj elemenata u promenljivu na koju pokazuje n a zatim i
 * cita toliko elemenata i smesta ih u niz array */
void read_float_array(double array[], int* n);

/* THREAD FUNCTION **********************************************************/

/* svaka nit dobija svoj deo za pretrazivanje (globalnog niza) a tek posle
 * pretrazivanja svog dela uporedjuje svoj lokalni maksimum sa globalnim */
void* thread_work(void* thread_arg) {
#ifdef PTW32_STATIC_LIB
	pthread_win32_thread_attach_np();
	{
#endif
	/* deklaracija lokalnih podataka niti */
	int thread_id = (intptr_t)thread_arg, ind, end;
	double local_max;

	/* nit ceka da se sve niti startuju */
	pthread_mutex_lock(&mutex);
	if (!thread_creation_finished)
		pthread_cond_wait(&thread_start_cond, &mutex);
	pthread_mutex_unlock(&mutex);

	/* inicijalizuje podatke za pretragu svog dela niza */
	ind = thread_id*array_slice;
	end = ind + array_slice;
	local_max = fabs(array[ind]);

	/* pretrazuje */
	for (ind++; ind<end; ind++)
		if (fabs(array[ind])>local_max)
			local_max = fabs(array[ind]);

	/* uporedjuje sa globalnim maksimumom */
	pthread_mutex_lock(&mutex);
	/* mora da saceka da glavna nit ispise prethodno promenjen maksimum */
	while (!thread_finished)
		pthread_cond_wait(&thread_finish_wait_cond, &mutex);
	if (local_max>max) {
		max = local_max;
		max_changed = 1;
	}
	thread_finished = 0;
	/* signalizira glavnoj niti da je zavrsila */
	pthread_cond_signal(&thread_finished_cond);
	pthread_mutex_unlock(&mutex);

	pthread_exit(NULL);

#ifdef PTW32_STATIC_LIB
}
pthread_win32_thread_detach_np();
#endif

	return NULL;
}

/* MAIN *********************************************************************/

int main(void) {
#ifdef PTW32_STATIC_LIB
	pthread_win32_process_attach_np();
	pthread_win32_thread_attach_np();
	{
#endif

	pthread_t threads[NUM_THREADS];
	pthread_attr_t attr;
	int t, rc;
	int left_to_work;

/* RESOURCE&DATA INITIALIZATION *********************************************/
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

    pthread_mutex_init(&mutex, NULL);
    pthread_cond_init(&thread_finished_cond, NULL);
    pthread_cond_init(&thread_finish_wait_cond, NULL);
    pthread_cond_init(&thread_start_cond, NULL);

    /* citanje niza sa ulaza i inicijalizacija odgovarajucih globalnih
     * promenljivih */
    read_float_array(array, &n);
    array_slice = n/NUM_THREADS;
    max = -1;

    /* inicijalizacija indikatora */
    thread_creation_finished = 0;
    thread_finished = 1;
    max_changed = 0;

/* THREAD CREATING **********************************************************/
	for (t = 0; t < NUM_THREADS; t++)
	{
	   rc = pthread_create(threads + t, &attr, thread_work, (void*)(intptr_t)t);
	   if (rc)
		   exit(EXIT_FAILURE);
	}

/* THREAD COLLABORATION *****************************************************/
	/* glavna nit signalizira ostalim da mogu da nastave sa izvrsavanjem */
	pthread_mutex_lock(&mutex);
	thread_creation_finished = 1;
	pthread_cond_broadcast(&thread_start_cond);
	pthread_mutex_unlock(&mutex);

	/* za svaku nit ceka potvrdu zavrsetka */
	for (left_to_work = NUM_THREADS; left_to_work>0; left_to_work--) {
		pthread_mutex_lock(&mutex);
		if (thread_finished)
			pthread_cond_wait(&thread_finished_cond, &mutex);
		/* ako je maksimum promenjen ispisi ga */
		if (max_changed) {
			printf("%lf\t", max);
			max_changed = 0;
		}
		thread_finished = 1;
		/* signaliziraj cekajucim nitima (ako takve postoje) da je poslednja
		 * promena zavrsena */
		pthread_cond_signal(&thread_finish_wait_cond);
		pthread_mutex_unlock(&mutex);
	}

	/* ispisi krajnju maksimalnu vrednost */
	printf("\n%lf", max);

/* RESOURCE RELEASING ******************************************************/
	pthread_attr_destroy(&attr);
	pthread_mutex_destroy(&mutex);
    pthread_cond_destroy(&thread_finished_cond);
    pthread_cond_destroy(&thread_finish_wait_cond);
    pthread_cond_destroy(&thread_start_cond);

    pthread_exit(NULL);

#ifdef PTW32_STATIC_LIB
	}
	pthread_win32_thread_detach_np();
	pthread_win32_process_detach_np();
#endif

	return EXIT_SUCCESS;
}

/* AUXILIARY FUNCTIONS (DEFINITIONS) ****************************************/

void read_float_array(double array[], int* n) {
	int i;
	scanf("%d", n);
	for (i=0; i<*n; i++)
		scanf("%lf",array+i);
}
